import * as angular from "angular";
import { isDeviceSecure, securityCheckResult } from "./device-security-check";

angular.module("mobile-app").service("deviceSecurityCheck", function ($rootScope, $ionicPopup, $timeout, $label) {
    /**
     * Popup to prompt user to set a passcode
     * and providing a button to check security settings again.
     *
     * @returns
     */
    function showTryAgainPopup(message: string) {
        const messageBody = message !== undefined ? message : $label("Device_Passcode_Setup_Prompt_Body");
        return $ionicPopup.show({
            title: `${$label("Device_Passcode_Setup_Prompt_Title")}<br/>${messageBody}`,
            cssClass: "login-confirm-popup",
            buttons: [{
                text: $label("Device_Passcode_Setup_Prompt_Button"),
                type: "button-positive",
            }
            ]
        });
    }

    /**
     * [Android Only]
     * Popup to notify user that to cache is disabled for the current OS
     * and recommend them to upgrade the OS.
     */
    function showUpgradeOSPopup() {
        return $ionicPopup.show({
            title: $label("Device_Security_Upgrade_OS_Notice"),
            cssClass: "login-confirm-popup",
            buttons: [{
                text: $label("Device_Security_Notice_Button"),
                type: "button-positive",
            }
            ]
        });
    }

    /**
     * Popup to notify user that the cache has been disabled
     * since it couldn't be determined if the device is secure.
     */
    function showErrorPopup() {
        return $ionicPopup.show({
            title: $label("Device_Security_Error_Notice"),
            cssClass: "login-confirm-popup",
            buttons: [{
                text: $label("Device_Security_Notice_Button"),
                type: "button-positive",
            }
            ]
        });
    }

    /**
     * Shows the appropriate popup for an error/indeterminate securityCheckResult.
     * Returns a promise which resolves when the user dismisses the popup.
     */
    this.showSecurityCheckError = function (result: securityCheckResult): Promise<void> {
        // if the cache isn't enabled due to security reasons,
        // inform the user and suggest upgrading the device
        if (result === securityCheckResult.indeterminate) {
            if (!upgradePopupDisplayed) {
                upgradePopupDisplayed = true;
                return showUpgradeOSPopup();
            }
        } else if (result === securityCheckResult.error) {
            return showErrorPopup();
        } else {
            return Promise.resolve();
        }
    };

    /**
     * Checks if the device has a security lock enabled.
     * waitUntilSecure:
     * if true, the function doesn't return unless:
     * * the device is secure
     * * the check failed due to an error
     * if false, the function returns after a single check with true/false
     *
     * @param {boolean} waitUntilSecure
     * @returns {Promise<securityCheckResult>} Resolves with one of the four valid states in securityCheckResult
     */
    function checkSecurityInternal(waitUntilSecure: boolean, message: string): Promise<securityCheckResult> {
        return isDeviceSecure()
            .then((result: securityCheckResult) => {
                if (result === securityCheckResult.notSecure && waitUntilSecure) {
                    return showTryAgainPopup(message).then(() => checkSecurityInternal(waitUntilSecure, message));
                }

                return result;
            })
            .catch((reason) => {
                const errorMessage = `Error encountered while checking for passcode/pin: ${reason}.`;
                console.error(errorMessage);
                return securityCheckResult.error;
            });
    }

    let securityCheckPromise;

    /**
     * Non-reentrant wrapper for checkSecurityInternal()
     *
     * @param {boolean} waitUntilSecure
     * @returns {Promise<securityCheckResult>}
     */
    this.checkSecurity = function (waitUntilSecure: boolean, message: string): Promise<securityCheckResult> {
        if (!securityCheckPromise) {
            securityCheckPromise = checkSecurityInternal(waitUntilSecure, message).then((result: securityCheckResult) => {
                securityCheckPromise = undefined;
                return result;
            });
        }
        return securityCheckPromise;
    };

    let upgradePopupDisplayed: boolean;
});