import * as q from "q";
import { trackEvent, events } from "../../../src/services/telemetry";
import { cacheManager } from "@ms-cds/data-model";
import { ISessionProvider } from "../../../src/connectors/ISessionProvider";

export enum securityCheckResult {
    notSecure,
    secure,
    indeterminate,
    error
}

/**
 * Parses the native security check result
 * and transforms it into one of the defined states
 *
 * @param {(string | boolean)} result
 * @returns {securityCheckResult}
 */
function transformSecurityCheckResult(result: string | boolean): securityCheckResult {
    if (result === false || result === "false") {
        return securityCheckResult.notSecure;
    }
    if (result === true || result === "true") {
        return securityCheckResult.secure;
    }
    if (result === "indeterminate") {
        return securityCheckResult.indeterminate;
    }

    return undefined;
}

/**
 *
 * Executes native code to check device security and
 * returns a promise which resolves with the result of the check
 * which is one of the four states of securityCheckResult
 * @returns Promise<securityCheckResult>
 */
export function isDeviceSecure(): Promise<securityCheckResult> {
    if (window.cordova) {
        return new Promise<securityCheckResult>((resolve, reject) => {
            if ((window as any).cordova && (window as any).cordova.plugins.SecurityPlugin && (window as any).cordova.plugins.SecurityPlugin.isDeviceSecure) {
                // we use a 'resolved' lock to make sure that
                // the security plugin and timeout function dont have a race condition
                let resolved = false;
                (window as any).cordova.plugins.SecurityPlugin.isDeviceSecure(function (result) {
                    if (!resolved) {
                        resolved = true;
                        const securityResult = transformSecurityCheckResult(result);

                        if (securityResult === securityCheckResult.indeterminate) {
                            trackEvent(events.securityCheckIncompatibleDevice.Name);
                        }

                        if (securityResult === securityCheckResult.notSecure) {
                            trackEvent(events.securityCheckDeviceNotSecure.Name);
                        }

                        resolve(securityResult);
                    }
                });
                // Must resolve within a given time (3 sec)
                setTimeout(function () {
                    if (!resolved) {
                        resolved = true;
                        console.error("Error encountered while checking for passcode/pin: Native code failure");
                        resolve(securityCheckResult.error);
                        trackEvent(events.securityCheckNativeError.Name);
                    }
                }, 3000);

            } else {
                console.error("Error encountered while checking for passcode/pin: Security Check Plugin not found");
                // plugin not loaded correctly / does not exist
                resolve(securityCheckResult.error);
                trackEvent(events.securityCheckPluginNotFound.Name);
            }
        });
    } else {
        // emulator
        return Promise.resolve(securityCheckResult.secure);
    }
}

let cacheDisabled: boolean;
/**
 * Clears and disables stored cache data
 * preserving login info
 * @param logUserOut If true, additionally logs the user out
 * @param sessionProvider the session provider that can be used to clear data
 */
export function disableDataCaching(logUserOut: boolean, sessionProvider: ISessionProvider) {
    if (cacheDisabled !== true) {
        if (logUserOut) {
            // clear cache and log user out
            sessionProvider.clearLogin(true);
        } else {
            // clear cache only
            sessionProvider.clearCache(true);
        }
        cacheManager.setStorageProvider(null);
        cacheDisabled = true;
    }
}

/**
 * Re-enables cache if not already enabled
 *
 * @export
 * @returns {q.Promise<any>} Resolves when cache has been successfully enabled
 */
export function enableDataCaching(): q.Promise<any> {
    if (cacheDisabled !== false) {
        cacheDisabled = false;
        return cacheManager.initializeFileStore();
    } else {
        return q.resolve<any>(undefined);
    }
}