import * as angular from "angular";
import { IEventHook } from "../../../src/Event/IEvent";

interface IInfiniteScrollScope extends angular.IScope {
    onInfinite: () => Promise<void> | undefined;
    distance: number;
    onContentChange: IEventHook<void>;
}

const closestAncestorWithClass = function (element: angular.IRootElementService, className) {
    let parent = element.parent();
    while (parent) {
        if (parent[0] && parent[0].classList && parent[0].classList.contains(className)) {
            return parent;
        }
        parent = parent.parent();
    }
    return null;
};

/**
 * Function to check if an element is a descendant of another element
 * @param parent
 * @param child
 * Returns true if the relationship is true, false if not
 */
const isDescendant = function (parent: HTMLElement, child: HTMLElement) {
    let immediateParent = child.parentNode;
    while (immediateParent) {
        if (immediateParent === parent) {
            return true;
        }
        immediateParent = immediateParent.parentNode;
    }
    return false;
};

/**
 * Infinite scroll triggers a callback when the bottom (a percentage of the client height) of a scroll region is reached.
 * The height calculation, however, is based off initial static height of the scroll region (initially and on a scroll event).
 * Thus, dynamic renders (using templates) do not contribute to the height calculation.
 */
angular.module("mobile-app").directive("infiniteScroll", function ($timeout) {
    return {
        restrict: "E",
        scope: {
            onInfinite: "&",
            distance: "=",
            onContentChange: "=?"
        },
        link: function (scope: IInfiniteScrollScope, element: angular.IRootElementService, attrs) {
            let scrollableParent = undefined;
            let parentScrollHeight = undefined;
            let parentClientHeight = undefined;
            let heightsDefined = false;

            let infiniteInProgress: boolean = false;

            const recalculateHeight = function () {
                if (scrollableParent) {
                    parentScrollHeight = scrollableParent.scrollHeight;
                    parentClientHeight = scrollableParent.clientHeight;
                }
            };

            let heightCalculationTimer: number = undefined;

            /**
             * This function only recalculates heights if they are not defined
             * If heightCalculationTimer is pending, we clear it
             * We set/reset heightCalculationTimer always
             * ensuring that we don't reset our flags (heightsDefined) until the timer times out
             */
            const tryRecalculateHeight = function () {
                // If a timer is pending, reset it
                if (heightCalculationTimer) {
                    clearTimeout(heightCalculationTimer);
                    heightCalculationTimer = undefined;
                }

                // If heights are not defined - recalculate them
                if (!heightsDefined) {
                    recalculateHeight();
                    heightsDefined = true;
                }

                heightCalculationTimer = setTimeout(() => {
                    parentClientHeight = undefined;
                    parentScrollHeight = undefined;
                    heightsDefined = false;
                    heightCalculationTimer = undefined;
                }, 50);
            };

            const afterInfinite = function () {
                scope.$apply();
                heightsDefined = false;
                infiniteInProgress = false;
            };

            const scrollListener = (event) => {
                // to do check descendant
                if (scrollableParent === undefined) {
                    if (isDescendant(event.target, element[0])) {
                        scrollableParent = event.target;
                    } else {
                        // scrolled parent is not an ancestor
                        return;
                    }
                }

                tryRecalculateHeight();
                const scrollTop = scrollableParent.scrollTop;
                const distanceToBottom = parentScrollHeight - (parentClientHeight + scrollTop);
                if (distanceToBottom / parentClientHeight < (scope.distance / 100) && !infiniteInProgress) {
                    infiniteInProgress = true;
                    let result = scope.onInfinite();
                    // ensure the result is resolved as a promise
                    let promise = Promise
                        .resolve()
                        .then(() => result);
                    promise
                        .then(afterInfinite)
                        .catch(afterInfinite);
                }
            };

            let timeoutPending = false;
            /**
             * If the content is smaller than the document height, trigger onInfinite callback
             */
            const processContentChange = () => {
                if (!timeoutPending) {
                    timeoutPending = true;
                    $timeout(() => {
                        timeoutPending = false;
                        if (element[0].parentElement) {
                            if ((document.documentElement.clientHeight) > element[0].parentElement.scrollHeight) {
                                scope.onInfinite();
                            }
                        }
                    });
                }
            };

            if (scope.onContentChange) {
                scope.onContentChange.subscribe(processContentChange);
            }

            // Process intial changes.
            processContentChange();

            window.addEventListener("scroll", scrollListener, true);

            // removing listeners on element destroy
            scope.$on("$destroy", function () {
                window.removeEventListener("scroll", scrollListener);
                scope.onContentChange.unsubscribe(processContentChange);
            });
        }
    };
});
