import * as angular from "angular";

interface IPullToRefreshScope extends angular.IScope {
    onRefresh: () => Promise<void> | undefined;
}

class State {
    private currentState: any;
    private transitionTimer: number;
    private pendingState: any;
    constructor(initialState) {
        this.currentState = initialState;
    }

    getState() {
        return this.currentState;
    }

    setState(newState) {
        // if a delayed transition is already happening
        if (this.transitionTimer !== undefined) {
            // prevent the older transition
            clearTimeout(this.transitionTimer);
            this.transitionTimer = undefined;
            this.pendingState = undefined;
        }

        this.currentState = newState;
    }

    /**
     * This is like setState but waits for the given delay(in ms)
     * before setting new state.
     * Note that if another delayed state change (for the same state) is requested before the timeout,
     * that request will be ignored.
     * To force a state change, call setState(newState) instead.
     *
     * @param {any} newState
     * @param {any} delay
     * @returns
     *
     * @memberOf State
     */
    setStateWithDelay(newState, delay: number) {
        // if a delayed transition is already happening
        if (this.transitionTimer !== undefined && this.pendingState === newState) {
            // do not allow other transitions to take place
            return;
        }

        this.pendingState = newState;
        this.transitionTimer = setTimeout(() => {
            this.currentState = newState;
            this.pendingState = undefined;
            this.transitionTimer = undefined;
        }, delay);
    }
}

angular.module("mobile-app").directive("pullToRefresh", function ($ionicGesture, $compile) {
    return {
        restrict: "A",
        scope: {
            onRefresh: "&",
        },
        link: function (scope: IPullToRefreshScope, element: angular.IRootElementService, attrs) {

            // the pull to refresh arrow
            const pullToRefreshArrow = angular.element('<div class="refresh-arrow"></div>');
            pullToRefreshArrow.css({
                position: "absolute",
                top: "-25px",
                border: "solid black",
                borderWidth: "0 .2em .2em 0",
                padding: ".20em",
                transform: "rotate(45deg)",
                width: "15px",
                height: "15px",
                left: "calc(50% - 10.6px)",
                transition: "transform 0.2s",
            });

            // ion spinner
            const spinnerElement = angular.element("<ion-spinner></ion-spinner>");
            $compile(spinnerElement)(scope);
            spinnerElement.css({
                left: "calc(50% - 17px)",
                position: "absolute",
                top: "-38px",
            });

            // hide the spinner svg initially by making height 0
            const refreshingSvg = spinnerElement.find("svg").eq(0);
            refreshingSvg.css({
                height: "0px"
            });

            element.prepend(pullToRefreshArrow);
            element.prepend(spinnerElement);

            const showElement = (ele) => {
                ele.css({
                    display: ""
                });
            };

            const hideElement = (ele) => {
                ele.css({
                    display: "none"
                });
            };

            // thresholds
            const initialTranslate = 0.0; // NOTE: This would result in undesirable behavior if the refreshing element is initially positioned by translating
            const refreshingTranslate = initialTranslate + 50; // position of the view during refresh
            const refreshThreshold = initialTranslate + 50; // minimum pull to trigger refresh
            const dragThreshold = initialTranslate + 150; // maximum possible drag

            /**
             * Function to move a dom element to the given position smoothly
             */
            const smoothReposition = function (ele, toPosition, delay) {
                setTimeout(() => {
                    ele.css("transition", "transform 0.2s");
                    ele.css("transform", `translate3d(0px, ${toPosition}px, 0px) scale(1)`);
                }, delay);
            };

            let affectedElements = [];
            let ptrSetup: boolean = false;

            /**
             * Function to unfreeze scroll behavior on affectedElements
             */
            const releaseAffectedElements = () => {
                affectedElements.forEach((ele) => {
                    ele.classList.remove("allow-scroll-isDisabled");
                });
                affectedElements = [];
            };

            /**
             * Helper ease-in curve function.
             * Domain (0,1) -> Range (0,1)
             */
            const easingFunction = (x: number) => {
                let y: number;
                x = x / 1.7;
                if (x >= 1) {
                    y = 1;
                } else if (x <= 0) {
                    y = 0;
                } else {
                    y = x;
                }
                return y;
            };

            // Set of possible states
            enum States {
                INACTIVE,
                SCROLLING,
                TRY_DRAG,
                DRAGGING,
                DRAG_RELEASE,
                REFRESHING
            }

            const state = new State(States.INACTIVE);
            let previousTranslate: number = initialTranslate;
            let currentTranslate: number = undefined;
            let initialDeltaY: number = undefined;

            /**
             * Function called during the DRAGGING state.
             * Moves the element in accordance to the drag gesture
             */
            const onDrag = (event) => {
                // setup pull to refresh only once per complete pull
                if (!ptrSetup) {
                    initialDeltaY = event.gesture.deltaY;
                    hideElement(spinnerElement);
                    showElement(pullToRefreshArrow);
                    pullToRefreshArrow.css({
                        transform: "rotate(45deg)"
                    });
                    element.css("transition", "");
                    let el = event.target as HTMLElement;
                    while (el.parentNode) {
                        if (el.classList && el.classList.contains("allow-scroll")) {
                            affectedElements.push(el);
                            el.classList.add("allow-scroll-isDisabled");
                        }
                        el = el.parentNode as HTMLElement;
                    }
                    ptrSetup = true;
                }

                const deltaY = event.gesture.deltaY - initialDeltaY;

                // move the element as a function of deltaY
                const deltaTranslate = easingFunction(deltaY / (dragThreshold)) * dragThreshold;
                element.css("transform", `translate3d(0px, ${deltaTranslate}px, 0px) scale(1)`);
                currentTranslate = deltaTranslate;

                // rotate the arrow to point up if beyond refreshThreshold
                if (previousTranslate <= refreshThreshold && currentTranslate >= refreshThreshold) {
                    pullToRefreshArrow.css({
                        transform: "rotate(-135deg)"
                    });
                } else if (previousTranslate >= refreshThreshold && currentTranslate <= refreshThreshold) {
                    pullToRefreshArrow.css({
                        transform: "rotate(45deg)"
                    });
                }

                previousTranslate = currentTranslate;
            };

            /**
             * Function called during the DRAG_RELEASE state.
             */
            const onDragRelease = (event) => {
                ptrSetup = false;
                initialDeltaY = undefined;
                releaseAffectedElements();

                if (currentTranslate > refreshThreshold) {
                    state.setState(States.REFRESHING);
                    smoothReposition(element, refreshingTranslate, 0);
                    hideElement(pullToRefreshArrow);
                    showElement(spinnerElement);

                    // reset refreshingSvg so it is visible
                    refreshingSvg.css({
                        transition: "",
                        height: ""
                    });

                    const afterRefresh = () => {
                        scope.$apply(() => {
                            // animate the refreshing element and pull up the view after a delay
                            smoothReposition(element, initialTranslate, 200);
                            refreshingSvg.css({
                                transition: "height ease-in 0.3s",
                                height: "0px"
                            });
                            state.setState(States.INACTIVE);
                        });
                    };

                    const refreshResult = scope.onRefresh();
                    // if onRefresh returns a promise
                    if (refreshResult && refreshResult.then) {
                        refreshResult.then(afterRefresh, afterRefresh).catch(afterRefresh);
                    } else {
                        // otherwise remove the refresh element after 1s
                        setTimeout(afterRefresh, 1000);
                    }

                } else {
                    smoothReposition(element, initialTranslate, 0);
                    state.setState(States.INACTIVE);
                }
            };

            /**
             * EVENT LISTENERS
             */
            // Scroll Event
            let scrollTimer = undefined;
            const scrollListener = (event) => {
                const scrollTop = (event.target as Element).scrollTop;
                const currentState = state.getState();
                if (scrollTop > 0) {
                    // dont change the refreshing state but allow scroll
                    if (currentState === States.INACTIVE || currentState === States.TRY_DRAG) {
                        state.setState(States.SCROLLING);
                    }

                    if (scrollTimer !== undefined) {
                        clearTimeout(scrollTimer);
                    }

                    scrollTimer = setTimeout(() => {
                        if (state.getState() === States.SCROLLING) {
                            state.setState(States.INACTIVE);
                        }
                        scrollTimer = undefined;
                    }, 10);
                } else {
                    // IOS specific case, bouncy scroll can cause negatative scroll offset on the top which implies dragging
                    if (currentState === States.TRY_DRAG) {
                        state.setState(States.DRAGGING);
                    }
                }
            };
            window.addEventListener("scroll", scrollListener, true);

            // Drag Event
            const dragListener = (event) => {
                const currentState = state.getState();
                switch (currentState) {
                    case States.INACTIVE:
                        // prevent dragging from triggering when dragged from the bottom of a list/page
                        if (event.gesture.direction === "down") {
                            state.setState(States.TRY_DRAG);
                        }
                        break;
                    case States.TRY_DRAG:
                        if (event.gesture.direction === "down") {
                            // allow 40ms to receive more scroll events
                            state.setStateWithDelay(States.DRAGGING, 40);
                        }
                        break;
                    case States.DRAGGING:
                        onDrag(event);
                    default: return; // Returns if any other state, including SCROLLING, DRAG_RELEASE
                }
            };
            const ionicDragGesture = $ionicGesture.on("drag", dragListener, element, { passive: true });

            // Touch Release Event
            const releaseListener = (event) => {
                if (state.getState() === States.DRAGGING) {
                    state.setState(States.DRAG_RELEASE);
                    onDragRelease(event);
                } else if (state.getState() === States.TRY_DRAG) {
                    state.setState(States.INACTIVE);
                }
            };
            const ionicReleaseGesture = $ionicGesture.on("release", releaseListener, element);

            // removing listeners on element destroy
            scope.$on("$destroy", function () {
                window.removeEventListener("scroll", scrollListener);
                $ionicGesture.off(ionicDragGesture, "drag", dragListener);
                $ionicGesture.off(ionicReleaseGesture, "release", releaseListener);
            });
        }
    };
});
