import { IPage, IPageOptions } from "../../src/view-model/IPage";
import { Page, IPageConfig, PageLoadState } from "../../src/view-model/page";
import * as angular from "angular";
import { PageLink } from "../../src/view-model/Control/PageLink/PageLink";

export interface IFormScope extends angular.IScope {
    stateParams: IPageOptions;
    pageInit: any;
    page: Page;
    $label: any;
    doRefresh: () => void;
    getState(): any;
    getErrorCode(): any;
    showActions: () => void;
    canShowActions: () => boolean;
    getVisibleActions: () => PageLink[];
    getPageClassnames: () => string;
    showErrorMessageDetails: (message: any) => void;
}

export interface IFormServices {
    linkFormState(scope: IFormScope, page: any): void;
    createForm: (args: IPageOptions, scope: IFormScope) => Page;
}

angular.module("mobile-app.services").factory("$page", function (
    $timeout,
    metadataServices,
    telemetry,
    $ionicActionSheet,
    $state,
    $label,
    helperServices,
    $navigate,
    businessLogic,
    networkStatus,
    dataServices,
    $accessibility,
    $ionicPopup,
    notificationServices,
    $ionicNavBarDelegate,
    cacheServices
) {
    return {
        linkFormState: function (scope: IFormScope, page) {
            // Track if user caused a refresh gesture
            let userRefreshing = false;

            let hasInfoBar: boolean = false;
            let infoBarAddRemoveTimer = null;
            let infoBarShownTimestamp = null;
            const infoBarMinVisibleTime = 2000; // minimum time info bar must be visible
            const infoBarLoadingDelay = 800;    // delay before 'Loading' infobar shows up
            const infoBarRefreshingDelay = 2000; // delay before 'Refreshing' infobar shows up

            scope.$watch(function () {
                return page.getLoadState();
            }, function (dataState) {

                if (infoBarAddRemoveTimer) {
                    $timeout.cancel(infoBarAddRemoveTimer);
                    infoBarAddRemoveTimer = null;
                }

                if (dataState === "error") {
                    // immediately set the info bar class
                    hasInfoBar = true;
                    infoBarShownTimestamp = Date.now();
                } else if (dataState === "refreshing" || dataState === "loading") {
                    // if pull to refresh or info bar is already set or page is showing complete data, do not do anything
                    if (!userRefreshing && !hasInfoBar && page.hasPartialData) {
                        // otherwise load them after a defined delay
                        infoBarAddRemoveTimer = $timeout(function () {
                            hasInfoBar = true;
                            infoBarShownTimestamp = Date.now();
                            infoBarAddRemoveTimer = null;
                        },
                            dataState === "loading" ? infoBarLoadingDelay : infoBarRefreshingDelay);
                    }
                } else if (dataState === "loaded") {
                    if (hasInfoBar) {
                        let infoBarRemoveDelay = 0;

                        if (infoBarShownTimestamp) {
                            // remove infobar at least after infoBarMinVisibility, or immediately if its already over
                            infoBarRemoveDelay = Math.max(infoBarMinVisibleTime - (Date.now() - infoBarShownTimestamp), 0);
                        }

                        // Remove the info bar after calculated delay
                        infoBarAddRemoveTimer = $timeout(function () {
                            hasInfoBar = false;
                            infoBarShownTimestamp = null;
                            infoBarAddRemoveTimer = null;
                        },
                            infoBarRemoveDelay);
                    }
                }
            });

            /**
             * This method provides class names to pages based on their states.
             *
             * Note: We add the pre-init css class on ion-view in both detail-page.html and task.html
             * eventhough we get it from getPageClassnames()
             * This is to avoid the flicker (preinit sets opacity to 0) that results on page load before ng - class adds the class
             * The class is removed eventually by ng-class assuming it tries adding it itself
             */
            scope.getPageClassnames = function () {
                let classes = [];

                if (hasInfoBar) {
                    classes.push("has-info-bar");
                }

                if (page.isInitializing()) {
                    classes.push("pre-init");
                }

                if (page.metadata.ThemeColorName) {
                    classes.push(page.metadata.ThemeColorName);
                }

                return classes.join(" ");
            };
            scope.doRefresh = () => {
                userRefreshing = true;
                return page.doRefresh().then(() => {
                    userRefreshing = false;
                    scope.$apply();
                }).catch(() => {
                    userRefreshing = false;
                    scope.$apply();
                });
            };
        },
        createForm: function (args: IPageOptions, scope: IFormScope): Page {
            const pageConfig: IPageConfig = {
                $dataService: dataServices,
                $metadataService: metadataServices,
                $cacheService: cacheServices,
                $navigate: $navigate,
                $businessLogic: businessLogic,
                $networkStatus: networkStatus,
                $telemetry: telemetry,
                $label: $label,
                $pageInit: scope.pageInit,
                onFormChanged: function () {
                    scope.$apply();
                },
                loadDelay: 300,
                $notificationService: notificationServices
            };

            scope.$label = $label;

            const page = new Page(args, pageConfig);

            page.metadataLoaded.then(() => {
                $accessibility.speak(page.title);
            });

            if (scope.page) {
                scope.page.close();
                scope.page = null;
            }

            scope.page = page;
            this.linkFormState(scope, page);

            page.onLoadStateChange.subscribe(() => $timeout(function () { return; }));

            scope.getState = function () {
                return page.getLoadState();
            };

            scope.getErrorCode = function () {
                return page.getLoadErrorCode();
            };

            scope.showErrorMessageDetails = function (message: any) {
                $ionicPopup.alert({
                    title: $label("ErrorDetails"),
                    subTitle: message.ExtendedMessage,
                    template: message.ActionDescription
                });
            };

            scope.canShowActions = function () {
                return scope.getVisibleActions().length > 0;
            };

            scope.getVisibleActions = function () {
                return page.getExecutableActions().filter((action) => {
                    return !action.hidden;
                });
            };

            scope.showActions = () => {
                let actions = scope.getVisibleActions();
                let buttons = actions.map(function (action) {
                    return { text: action.metadata().Label };
                });
                let pageMetadata = page.getPageMetadata();
                let workSpaceNameProperty = telemetry.events.actionsShown.TelemetryProperties.workspaceName.Name;
                let formNameProperty = telemetry.events.actionsShown.TelemetryProperties.formName.Name;
                let pageNameProperty = telemetry.events.actionsShown.TelemetryProperties.pageName.Name;
                let actionCountPropertyName = telemetry.events.actionsShown.TelemetryProperties.actionCount.Name;
                let formName = pageMetadata ? pageMetadata.MasterForm : "";
                let pageName = pageMetadata ? pageMetadata.Title : "";
                let actionCount = actions && actions.length ? actions.length : 0;
                let workspaceName = cacheServices.getData("$workspaceNameForTelemetry");
                telemetry.trackEvent(telemetry.events.actionsShown.Name, {
                    [workSpaceNameProperty]: workspaceName, [formNameProperty]: formName, [pageNameProperty]: pageName, [actionCountPropertyName]: actionCount
                });
                $ionicActionSheet.show({
                    titleText: $label("Actions_label"),
                    cancelText: $label("Actions_cancel"),
                    buttons: buttons,
                    buttonClicked: (buttonIndex) => {
                        let pagelinkCtrl = actions[buttonIndex];
                        let actionMetadata = pagelinkCtrl.metadata();
                        let navigation = pagelinkCtrl.getNavigationHandler();

                        if (actionMetadata.OnNavigate) {
                            actionMetadata.OnNavigate(navigation);
                        }

                        if (navigation.to) {
                            $navigate(navigation);
                        }

                        return true;
                    }
                });
            };

            let resume = () => {
                page.resume().then(() => {
                    $accessibility.speak(page.title);
                    scope.$apply();
                }).catch(() => {
                    scope.$apply();
                });
            };

            // Wire up view events
            scope.$on("$ionicView.unloaded", page.close);
            scope.$on("$ionicView.enter", resume);
            scope.$on("$ionicView.leave", page.suspend);

            return page;
        },
    };
});