import * as angular from "angular";
import { networkStatusProvider, DataFilter, DataFilterOperator } from "@ms-cds/data-model";
import * as localization from "../../src/localization/localization";
import * as telemetry from "../../src/services/telemetry";
import { IConnector } from "../../src/connectors/IConnector";
import { AxConnector } from "../../src/connectors/ax/connector";
let RuntimeServices = require("@ax/runtime-service");
import { businessLogic } from "../../src/services/businessLogic";
import { queryWalker, SyncUnitStatus, cacheManager, newGuid, arraySort } from "@ms-cds/data-model";
import { EntityRef } from "@ms-cds/data-contracts";
import * as controlBinding from "../../src/view-model/controlBinding";
import { ICamera } from "../../src/view-model/Control/FileUploader/FileUploader";
import { isDeviceSecure, securityCheckResult, enableDataCaching, disableDataCaching } from "../components/device-security-check/device-security-check";

export function registerApp() {
    angular.module("mobile-app", ["ionic", "mobile-app.controllers", "mobile-app.services", "ngCordova", "app"]);
    angular.module("mobile-app.controllers", []);
    angular.module("mobile-app.services", []);
    angular.module("app", []);
}

// BEGIN References to be removed
let axSession = require("@ax/connectors/ax/ax-session");
// END References to be removed

const navigateToRoot = function ($ionicHistory, $state, navigation) {
    if ($ionicHistory && $state && navigation) {
        let history = $ionicHistory.viewHistory().histories[$ionicHistory.currentHistoryId()];
        let historyCount: number = 0;
        if (history) {
            historyCount = history.stack.length - 1;
        }

        let goBackCount: number = 0;
        if (historyCount > 0) {
            if (navigation.to.toLowerCase() === "home") {
                goBackCount = ((-1) * historyCount) + 1;
            } else {
                goBackCount = ((-1) * historyCount);
            }
        }

        if (goBackCount < 0) {
            $ionicHistory.goBack(goBackCount);
        }
    }
};

/**
 * Initializes the app
 */
export function initializeApp(checkSecurity: boolean): Promise<void> {
    // Initialize the telemetry channel.
    telemetry.start();

    let connector = AxConnector.create();
    let runtimeServices = RuntimeServices.create(connector);
    // Make the declarative runtime services available to components which are not properly factored
    (window as any).runtimeServices = runtimeServices;

    // Initialize the language for the app
    localization.setLanguage(navigator.language.toLowerCase());
    connector.setLanguage(localization.getLanguage());

    // Register all modules
    registerModules(connector, runtimeServices);

    if (checkSecurity) {
        telemetry.trackEvent(telemetry.events.deviceSecurityCheckStarted.Name);
        return isDeviceSecure().then(result => {
            telemetry.trackEvent(telemetry.events.deviceSecurityCheckEnded.Name);
            if (result !== securityCheckResult.secure) {
                if (result === securityCheckResult.notSecure) {
                    // silently disable and clear cache and log the user out
                    disableDataCaching(true, connector);
                } else {
                    // incompatible device
                    // disable and clear cache and not log the user out
                    disableDataCaching(false, connector);

                    // register a run listener to show the appropriate popup
                    angular.module("mobile-app")
                        .run(function (deviceSecurityCheck) {
                            deviceSecurityCheck.showSecurityCheckError(result);
                        });
                }
            }
        });
    } else {
        return Promise.resolve();
    }
}

function registerModules(connector: IConnector, runtimeServices) {
    angular.module("app").run(function ($state, $ionicHistory, $ionicPlatform, $label, $ionicPopup) {
        $ionicPlatform.registerBackButtonAction(function (event) {
            if ($state.current.disableHardwareBackButton) {
                event.preventDefault();
            } else {
                $ionicHistory.goBack();
            }
        }, 501);
    });

    // Define additional modules
    // All communication directly with AX endpoint should go through this service-
    angular.module("mobile-app.services").factory("axSession", function () {
        return axSession;
    });

    // Connector used in the app is exposed as a depdendency
    angular.module("mobile-app.services").factory("connector", function () {
        return connector;
    });

    angular.module("mobile-app.services").factory("cacheManager", function () {
        return cacheManager;
    });

    angular.module("mobile-app.services").factory("helperServices", function () {
        return { arraySort };
    });

    angular.module("mobile-app.services").factory("networkStatus", function () {
        return networkStatusProvider;
    });

    angular.module("mobile-app.services").factory("telemetry", function () {
        return telemetry;
    });

    angular.module("mobile-app.services").factory("dataServices", function () {
        return runtimeServices.dataServices;
    });

    angular.module("mobile-app.services").factory("metadataServices", function () {
        return runtimeServices.metadataServices;
    });

    angular.module("mobile-app.services").factory("cacheServices", function () {
        return runtimeServices.cacheServices;
    });

    angular.module("mobile-app.services").factory("notificationServices", function () {
        return runtimeServices.notificationServices;
    });

    angular.module("mobile-app.services").factory("businessLogic", function () {
        return businessLogic;
    });

    // TODO: Refactor usage of clone data from task.js
    angular.module("mobile-app.services").factory("queryWalker", function () {
        return queryWalker;
    });

    angular.module("mobile-app.services").factory("EntityRef", function () {
        return EntityRef;
    });

    angular.module("mobile-app.services").factory("SyncUnitStatus", function () {
        return SyncUnitStatus;
    });

    angular.module("mobile-app.services").factory("DataFilter", function () {
        return DataFilter;
    });

    angular.module("mobile-app.services").factory("DataFilterOperator", function () {
        return DataFilterOperator;
    });

    angular.module("mobile-app.services").factory("newGuid", function () {
        return newGuid;
    });

    angular.module("mobile-app.services").factory("$label", function () {
        return localization.resolveLabel;
    });

    angular.module("mobile-app.services").factory("controlBinding", function () {
        return controlBinding;
    });

    let $navigate;

    angular.module("mobile-app.services").factory("$navigate", function ($state, $ionicHistory) {
        //TODO: We should move this to seperate file and call that method from here.
        if (!$navigate) {
            // create the instance if the navigator function
            $navigate = function (navigation) {
                if (navigation) {
                    if (navigation.url) {
                        //_system is to open in system web browser
                        window.open(navigation.url, "_system");
                    } else if (navigation.to) {
                        if (navigation.to.toLowerCase() === "home" || navigation.to.toLowerCase() === "exit") {
                            navigateToRoot($ionicHistory, $state, navigation);
                            return;
                        } else if (navigation.replace) {
                            // This removes current view firing navigation from navigation history stack
                            $ionicHistory.currentView($ionicHistory.backView());
                        }
                        // Now go where we are requested to
                        $state.go(navigation.to, navigation.params, navigation.options);
                    }
                }
            };
        }
        return $navigate;
    });

    angular.module("mobile-app.services").factory("$accessibility", function ($timeout, $rootScope) {
        let isSrOn = false;
        if ((window as any).MobileAccessibility) {
            (window as any).MobileAccessibility.isScreenReaderRunning(function (mobAccOn) {
                $timeout(function () {
                    if (mobAccOn) {
                        isSrOn = $rootScope.srOn = true;
                    } else {
                        isSrOn = $rootScope.srOn = false;
                    }
                });
            });
        }

        let $accessibility = {
            isScreenReaderOn: function () {
                return isSrOn;
            },
            speak: function (text) {
                if ((window as any).MobileAccessibility) {
                    (window as any).MobileAccessibility.speak(text);
                }
            },
            deferredSpeak: function (text) {
                // Reads text without interruption of current screen reader's speaking.
                // If screen reader is not finished current phrase, pronounces text immediately after it
                // Otherwise reads text without a delay
                if ((window as any).MobileAccessibility) {
                    (window as any).MobileAccessibility.speak(text, 1);
                }
            },
        };

        return $accessibility;
    });
    angular.module("mobile-app.services").factory("$camera", function ($cordovaCamera) {
        let $camera: ICamera = {
            hasCameraPlugin: function () {
                return (window as any).cordova && $cordovaCamera;
            },
            getImage(options): Promise<string> {
                return $cordovaCamera.getPicture(options);

            },
            resolveLocalFileSystemURL: function (imagePath: string, success: EntryCallback, reject: EntryCallback): void {
                return window.resolveLocalFileSystemURL(imagePath, success, reject);
            }
        };
        return $camera;
    });
}